import os

from astropy.coordinates import SkyCoord
from astropy.table import Table, QTable
from astropy.tests.helper import assert_quantity_allclose
from astropy.io import fits
import astropy.units as u
import numpy as np
import pytest


def _make_catalog(with_units=u.deg, as_skycoord=False):
    """
    Create a sample catalog table with optional units and SkyCoord columns
    for testing purposes.
    """

    ra = [9.423, 9.421, 9.415] * (u.deg if with_units else 1)
    dec = [-33.711, -33.716, -33.717] * (u.deg if with_units else 1)
    obj_id = ['source1', 'souce2', 'source3']
    flux = [10, 20, 30] * (u.Jy if with_units else 1)

    tab_cls = QTable if with_units else Table

    if as_skycoord:
        sc = SkyCoord(ra, dec)
        return tab_cls(data=[sc, obj_id, flux], names=['SkyCoord', 'Obj_ID', 'flux'])
    return tab_cls(data=[ra, dec, obj_id, flux],
                   names=['RA', 'Dec', 'Obj_ID', 'flux'])


@pytest.mark.parametrize("from_file", [True, False])
@pytest.mark.parametrize("with_units", [True, False])
def test_load_catalog(imviz_helper, tmp_path, from_file, with_units):
    """
    Verify the basic functionality loading catalogs into the data collection.
    Test cases cover both in-memory Astropy tables and ECSV files, and catalogs
    with and without units on RA/Dec columns already assigned.
    """
    imviz_helper.app.state.catalogs_in_dc = True

    # basic catalog with RA, Dec, ID, and Flux columns and units
    catalog_obj = _make_catalog(with_units)

    if from_file:
        fn = os.path.join(tmp_path, "catalog.ecsv")
        catalog_obj.write(fn)
        catalog = fn
    else:
        catalog = catalog_obj

    # load catalog
    imviz_helper.load(catalog)

    # ensure that it is in the data collection with the correct label "Catalog"
    dc = imviz_helper.app.data_collection
    assert len(dc) == 1
    assert 'Catalog' in imviz_helper.app.data_collection.labels

    # make sure 'RA' column was renamed to Right Ascension and 'Dec' to 'Declination'
    # in the data collection for consistency, and that the table in the data
    # collection always has units
    qtab = imviz_helper.app.data_collection[0].get_object(QTable)
    assert 'Right Ascension' in qtab.colnames
    assert 'Declination' in qtab.colnames
    # make sure only ra and dec loaded, since we didn't specify more columns
    assert len(qtab.colnames) == 2
    # and that it has the correct contents, and always has units assigned
    # when data is loaded from a unitless table, units should always be assigned
    # to the catalog in the data collection based on selections in the loader
    un = 1
    if not with_units:
        un = u.deg
    assert_quantity_allclose(qtab['Right Ascension'], catalog_obj['RA'] * un)
    assert_quantity_allclose(qtab['Declination'], catalog_obj['Dec'] * un)
    assert qtab['Right Ascension'].unit == qtab['Declination'].unit == u.deg

    # make sure 'col_ra_has_unit' is True if the input catalog has units, and is
    # False if the input doesn't have units. This controls the dropdown to
    # select unit is exposed, which is eventually assigned to the table column
    # in the data collection
    ldr = imviz_helper.loaders['file' if from_file else 'object']
    setattr(ldr, 'filepath' if from_file else 'object', catalog)
    assert ldr.importer._obj.col_ra_has_unit == with_units

    # load it again, make sure label incremented by 1
    imviz_helper.load(catalog)
    assert len(dc) == 2
    assert 'Catalog (1)' in imviz_helper.app.data_collection.labels

    # load with custom label and check label
    imviz_helper.load(catalog, data_label='my_catalog')
    assert len(dc) == 3
    assert 'my_catalog' in imviz_helper.app.data_collection.labels

    # test other loader API options. switch RA and Dec col just to test
    # non-default column selection
    imviz_helper.load(catalog, data_label='with_flux', col_other='flux',
                      col_ra='Dec', col_dec='RA')
    assert len(dc) == 4
    assert 'flux' in dc['with_flux'].get_object(QTable).colnames
    qtab = imviz_helper.app.data_collection[-1].get_object(QTable)
    assert_quantity_allclose(qtab['Right Ascension'], catalog_obj['Dec'] * un)
    assert_quantity_allclose(qtab['Declination'], catalog_obj['RA'] * un)


@pytest.mark.parametrize("from_file", [True, False])
def test_load_catalog_skycoord(imviz_helper, tmp_path, from_file):
    """
    Test loading a catalog with a SkyCoord column into the Imviz helper.

    This test verifies that when a catalog containing a SkyCoord column is
    loaded, the loader correctly uses this column as the default for
    Right Ascension (RA) and Declination (Dec).
    """
    # test loading catalog with SkyCoord column, which should be used
    # as default for Ra and Dec

    imviz_helper.app.state.catalogs_in_dc = True

    catalog_obj = _make_catalog(as_skycoord=True)

    if from_file:
        fn = os.path.join(tmp_path, "catalog.ecsv")
        catalog_obj.write(fn)
        catalog = fn
    else:
        catalog = catalog_obj

    # load catalog
    imviz_helper.load(catalog)

    dc = imviz_helper.app.data_collection
    assert len(dc) == 1
    assert 'Catalog' in imviz_helper.app.data_collection.labels

    # make sure 'RA' column was renamed to Right Ascension and 'Dec' to 'Declination'
    # in the data collection for consistency, and that the table in the data
    # collection always has units
    qtab = imviz_helper.app.data_collection[0].get_object(QTable)
    assert 'Right Ascension' in qtab.colnames
    assert 'Declination' in qtab.colnames
    # make sure only ra and dec loaded, since we didn't specify more columns
    assert len(qtab.colnames) == 2
    # and that it has the correct contents, and always has units assigned
    # when data is loaded from a unitless table, units should always be assigned
    # to the catalog in the data collection based on selections in the loader
    assert_quantity_allclose(qtab['Right Ascension'], catalog_obj['SkyCoord'].ra)
    assert_quantity_allclose(qtab['Declination'], catalog_obj['SkyCoord'].dec)


def test_invalid(imviz_helper, tmp_path):

    imviz_helper.app.state.catalogs_in_dc = True

    # make sure you can't load an empty table
    empty_table = Table()

    # empty table from object
    ldr = imviz_helper.loaders['object']
    ldr.object = empty_table
    assert 'Catalog' not in ldr.format.choices

    # empty table from file
    fn = os.path.join(tmp_path, "empty_catalog.ecsv")
    empty_table.write(fn)

    ldr = imviz_helper.loaders['file']
    ldr.filepath = fn
    assert 'Catalog' not in ldr.format.choices

    # while a fits file can be opened with table.read, it should not be
    # validated as a Catalog
    ldr = imviz_helper.loaders['object']
    ldr.object = fits.ImageHDU(np.ones((32, 25)))
    assert 'Image' in ldr.format.choices
    assert 'Catalog' not in ldr.format.choices



from astropy.coordinates import SkyCoord
from astropy.table import Table, QTable
import astropy.units as u
import numpy as np
from traitlets import Bool, List, Unicode, observe

from jdaviz.core.loaders.importers import BaseImporterToDataCollection
from jdaviz.core.template_mixin import SelectPluginComponent
from jdaviz.core.registries import loader_importer_registry
from jdaviz.core.user_api import ImporterUserApi

__all__ = ['CatalogImporter']


@loader_importer_registry("Catalog")
class CatalogImporter(BaseImporterToDataCollection):

    template_file = __file__, "./catalog.vue"

    col_ra_items = List().tag(sync=True)
    col_ra_selected = Unicode().tag(sync=True)

    col_dec_items = List().tag(sync=True)
    col_dec_selected = Unicode().tag(sync=True)

    col_ra_has_unit = Bool().tag(sync=True)  # if input already has units
    col_ra_unit_items = List().tag(sync=True)
    col_ra_unit_selected = Unicode().tag(sync=True)

    col_dec_has_unit = Bool().tag(sync=True)  # if input already has units
    col_dec_unit_items = List().tag(sync=True)
    col_dec_unit_selected = Unicode().tag(sync=True)

    # for non- ra and dec columns that the user wants to load
    col_other_items = List().tag(sync=True)
    col_other_selected = List().tag(sync=True)
    col_other_multiselect = Bool(True).tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not self.is_valid:
            return

        self.col_ra = SelectPluginComponent(self,
                                            items='col_ra_items',
                                            selected='col_ra_selected',
                                            manual_options=self._guess_ra_dec_cols('ra'))
        self.col_dec = SelectPluginComponent(self,
                                             items='col_dec_items',
                                             selected='col_dec_selected',
                                             manual_options=self._guess_ra_dec_cols('dec'))

        # making these a dropdown for now with some hard-coded unit choices
        # this could either be more flexible with a text input field which
        # is then validated as a unit, or a note to use a quantity table
        # if you want other units
        self.col_ra_unit = SelectPluginComponent(self,
                                                 items='col_ra_unit_items',
                                                 selected='col_ra_unit_selected',
                                                 manual_options=self._valid_coord_units('ra'))
        self.col_dec_unit = SelectPluginComponent(self,
                                                  items='col_dec_unit_items',
                                                  selected='col_dec_unit_selected',
                                                  manual_options=self._valid_coord_units('dec'))

        self.col_other = SelectPluginComponent(self,
                                               items='col_other_items',
                                               selected='col_other_selected',
                                               manual_options=self.input.colnames,
                                               multiselect='col_other_multiselect')

    def _guess_ra_dec_cols(self, col):
        """
        Rough guess at detecting ra, dec columns from input table by checking
        for the presence of a SkyCoord column, and if none exists then checking
        against some common source catalog column names, to determine initial
        selection for the column select dropdown. If no good candidate is found,
        the initial selection in the drop down for ra, dec columns will be the
        0th and 1stcolumns , respectivley.
        """

        tab = self.input
        colnames = self.input.colnames

        if colnames is None:
            return

        col_is_sc = [isinstance(tab[colnames[i]], SkyCoord) for i in range(len(colnames))]

        if np.any(col_is_sc):
            idx = np.where(col_is_sc)[0][0]

        else:
            all_column_names = np.array(x.lower() for x in colnames)
            get_idx = lambda x, s, d: np.where(np.isin(x, s))[0][0] if np.any(np.isin(x, s)) else d  # noqa

            if col == 'ra':
                col_possibilities = ['right ascension', 'ra', 'ra_deg', 'radeg',
                                     'radegrees', 'right ascension (degrees)',
                                     'ra_obj', 'raj2000', 'ra2000']
                idx = get_idx(all_column_names, col_possibilities, 0)
            elif col == 'dec':
                col_possibilities = ['declination', 'dec', 'dec_deg', 'decdeg',
                                     'decdegrees', 'declination (degrees)',
                                     'dec_obj', 'obj_dec', 'decj2000', 'dec2000']
                idx = get_idx(all_column_names, col_possibilities, 1)

        return colnames if idx == 0 else (colnames[idx:] + colnames[:idx])

    def _valid_coord_units(self, coord):
        """Valid choices for Ra, Dec units."""

        choices = ['deg', 'rad', 'arcmin', 'arcsec']

        if coord == 'ra':
            choices += ['hour minute second']
        elif coord == 'dec':
            choices += ['degree arcmin arcsec']

        return choices

    @observe('col_ra_selected')
    def _on_ra_col_selected(self, msg):
        """Check if the selected 'ra' column has units assigned already"""

        ra = self.col_ra_selected
        dec = self.col_dec_selected

        if ra == dec and not isinstance(self.input[ra], SkyCoord):
            self.resolver.import_disabled = True
        else:
            self.resolver.import_disabled = False

        has_units = False
        if hasattr(self.input[ra], 'unit'):
            if self.input[ra].unit is not None:
                has_units = True
                # ra column unit must be an angle unit
                if self.input[ra].unit.physical_type != 'angle':
                    has_units = False

        elif isinstance(self.input[self.col_ra_selected], SkyCoord):
            has_units = True

        if ra == dec and not isinstance(self.input[dec], SkyCoord):
            self.import_disabled = True
        else:
            self.import_disabled = False

        self.col_ra_has_unit = has_units

    @observe('col_dec_selected')
    def _on_dec_col_selected(self, msg):
        """Check if the selected 'dec' column has units assigned already"""

        ra = self.col_ra_selected
        dec = self.col_dec_selected

        if ra == dec:
            self.import_disabled = False

        has_units = False
        if hasattr(self.input[dec], 'unit'):
            if self.input[dec].unit is not None:
                has_units = True
                # dec column unit must be an angle unit
                if self.input[dec].unit.physical_type != 'angle':
                    has_units = False

        elif isinstance(self.input[dec], SkyCoord):
            has_units = True

        if ra == dec and not isinstance(self.input[dec], SkyCoord):
            self.import_disabled = True
        else:
            self.import_disabled = False

        self.col_dec_has_unit = has_units

    @staticmethod
    def _get_supported_viewers():
        return [{'label': 'Image', 'reference': 'imviz-image-viewer'}]

    @property
    def user_api(self):
        expose = ['col_ra', 'col_dec', 'col_other']
        return ImporterUserApi(self, expose=expose)

    @property
    def is_valid(self):
        if self.app.state.catalogs_in_dc is False:
            return False
        if self.app.config not in ('deconfigged', 'imviz', 'mastviz'):
            # NOTE: temporary during deconfig process
            return False
        if isinstance(self.input, (Table, QTable)) and len(self.input):
            return True
        return False

    @property
    def output_cols(self):

        # we will always have RA and Dec
        cols_all = [self.col_ra_selected, self.col_dec_selected] + self.col_other_selected

        return [col for col in set(cols_all) if col in self.input.colnames]

    @property
    def output(self):

        if (self.col_ra_selected not in self.input.colnames) or (self.col_dec_selected not in self.input.colnames):  # noqa
            return

        table = self.input[self.output_cols]
        output_table = QTable()

        # rename RA and Dec columns so that table in data collection always has
        # the same RA, Dec column names internally, add and units if they weren't
        # loaded in with units assigned (QTable)
        ra = None
        dec = None
        if isinstance(self.input[self.col_ra_selected], SkyCoord):
            ra = self.input[self.col_ra_selected].ra
        if isinstance(self.input[self.col_dec_selected], SkyCoord):
            dec = self.input[self.col_dec_selected].dec

        if ra is not None:
            output_table['Right Ascension'] = ra
        else:
            output_table['Right Ascension'] = table[self.col_ra_selected]
            # add units to ra if they weren't loaded in with units assigned
            if not self.col_ra_has_unit:
                output_table['Right Ascension'] *= u.Unit(self.col_ra_unit_selected)
        if dec is not None:
            output_table['Declination'] = dec
        else:
            output_table['Declination'] = table[self.col_dec_selected]
            if not self.col_dec_has_unit:
                output_table['Declination'] *= u.Unit(self.col_dec_unit_selected)

        # add additional columns to output table
        for col in self.output_cols:
            if col not in output_table.colnames + [self.col_ra_selected, self.col_dec_selected]:
                output_table[col] = table[col]

        return output_table



from jdaviz.app import Application
from jdaviz.core.loaders.resolvers.resolver import BaseResolver


# Create a minimal test class that mimics the resolver behavior
# without this we get an error when attempting to use BaseResolver directly
class TestBaseResolver(BaseResolver):
    template = ''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


def test_server_is_remote_callback():
    # Create app instance
    app = Application()

    # Test the sync
    test_obj = TestBaseResolver(app=app)
    settings = app.state.settings

    # Check default
    assert settings.get('server_is_remote') is False
    assert settings.get('server_is_remote') == test_obj.server_is_remote

    settings['server_is_remote'] = True
    assert settings.get('server_is_remote') == test_obj.server_is_remote

    # Ensure setting test_obj.server_is_remote does not backpropagate
    # (this behavior could change)
    test_obj.server_is_remote = False
    assert settings.get('server_is_remote') != test_obj.server_is_remote



import itertools
import numpy as np
from copy import deepcopy
import warnings

from astropy.nddata import StdDevUncertainty
from specutils import Spectrum, SpectrumList, SpectrumCollection
from traitlets import List, Bool, Any, observe

from jdaviz.core.unit_conversion_utils import (to_flux_density_unit,
                                               spectrum_ensure_flux_density_unit)
from jdaviz.core.registries import loader_importer_registry
from jdaviz.core.loaders.importers import (BaseImporterToDataCollection,
                                           _spectrum_assign_component_type)
from jdaviz.core.template_mixin import SelectFileExtensionComponent
from jdaviz.core.user_api import ImporterUserApi
from jdaviz.core.events import SnackbarMessage


__all__ = ['SpectrumListImporter', 'SpectrumListConcatenatedImporter']


@loader_importer_registry('1D Spectrum List')
class SpectrumListImporter(BaseImporterToDataCollection):
    template_file = __file__, "spectrum_list.vue"

    sources_items = List().tag(sync=True)
    sources_selected = Any().tag(sync=True)
    sources_multiselect = Bool(True).tag(sync=True)

    input_in_sb = Bool(False).tag(sync=True)
    convert_to_flux_density = Bool(True).tag(sync=True)

    disable_dropdown = Bool(False).tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.default_data_label_from_resolver:
            self.data_label_default = self.default_data_label_from_resolver
        elif self.app.config == 'specviz':
            self.data_label_default = '1D Spectrum'
        else:
            self.data_label_default = '1D Spectrum'

        sources_options = []

        if isinstance(self.input, Spectrum):
            speclist_input = SpectrumList(self.input_to_list_of_spec(self.input))
        else:
            speclist_input = self.input

        for index, spec in enumerate(speclist_input):
            if self.is_wfssmulti(spec):
                # ver, name are stand-ins for exposure and source_id
                # ver == exposure, name == source_id
                ver, name = self._extract_exposure_sourceid(spec)
                exposure_label = f"Exposure {ver}"

                label = f"{exposure_label}, Source ID: {name}"
                # Flipping the two from the variable naming convention
                name_ver = f"{ver}_{name}"
                suffix = f"EXP-{ver}_ID-{name}"

            else:
                name_ver = index
                name = index
                ver = index
                label = f"1D Spectrum at index: {index}"
                suffix = f"index-{index}"

            sources_options.append({'label': label,
                                    'index': index,
                                    'name': str(name),
                                    'ver': str(ver),
                                    'name_ver': str(name_ver),
                                    'suffix': suffix,
                                    'obj': self._apply_spectral_mask(spec)})

        self.sources = SelectFileExtensionComponent(self,
                                                    items='sources_items',
                                                    selected='sources_selected',
                                                    multiselect='sources_multiselect',
                                                    manual_options=sources_options)

        self.sources.selected = [self.sources.choices[0]]
        self._sources_items_helper = deepcopy(self.sources.items)

        # TODO: This observer will likely be removed in follow-up effort
        # If the resolver format is set to "1D Spectrum List", then we
        # only enable the import button if at least one spectrum is selected.
        self.resolver.observe(self._on_format_selected_change, names='format_selected')

    def _apply_kwargs(self, kwargs):
        applied_kwargs = super()._apply_kwargs(kwargs)
        if 'sources' not in applied_kwargs:
            msg_str = (f"The default source selection ({self.sources.selected}) will be used.\n"
                       f"To load additional sources, please specify them via dropdown or "
                       f"as follows:\n'{self.config}.load(filename, sources = [...]).")
            msg = SnackbarMessage(msg_str, color='warning', sender=self, timeout=10000)
            self.app.hub.broadcast(msg)
            warnings.warn(msg_str)
        return applied_kwargs

    @staticmethod
    def _get_supported_viewers():
        return [{'label': '1D Spectrum', 'reference': 'spectrum-1d-viewer'}]

    @property
    def user_api(self):
        expose = ['sources', 'convert_to_flux_density']
        return ImporterUserApi(self, expose)

    @property
    def is_valid(self):
        if self.app.config not in ('deconfigged', 'specviz'):
            # NOTE: temporary during deconfig process
            return False
        # TODO: should this be split into two loaders?
        # should a loader take a single input type, output a single output type,
        # or just have a consistent data_label and viewer?

        # If the input is a SpectrumList or SpectrumCollection, it
        # must be non-empty.
        if isinstance(self.input, (SpectrumList, SpectrumCollection)):
            return len(self.input) > 0

        return self._is_2d_spectrum

    @observe('sources_selected')
    def _on_sources_selected_change(self, change={}):
        if len(self.sources_selected) == 0:
            self.import_disabled = True
        else:
            self.import_disabled = False

            self.input_in_sb = bool(np.any([sp.flux.unit.physical_type == 'surface brightness'
                                            for sp in self.sources.selected_obj]))

    def _on_format_selected_change(self, change={}):
        if change['new'] == '1D Spectrum List':
            # Only perform this check if the selected importer is SpectrumListImporter
            # Otherwise other valid importers (in the case of Spectrum2D)
            # will also run and reset import_disabled
            if 'SpectrumListImporter' in str(type(self)).split('.')[-1]:
                self._on_sources_selected_change()

        elif change['new'] == '1D Spectrum Concatenated':
            # 2D Spectra load all for concatenated
            if self._is_2d_spectrum:
                self.import_disabled = False
            else:
                self._on_sources_selected_change()

        else:
            self.import_disabled = False

    def input_to_list_of_spec(self, inp):

        def this_row(field, i):
            if field is None:
                return None
            return field[i, :]

        if isinstance(inp, Spectrum):
            if inp.flux.ndim == 1:
                return [self._apply_spectral_mask(inp)]

            return [Spectrum(spectral_axis=inp.spectral_axis,
                             flux=this_row(inp.flux, i),
                             uncertainty=this_row(inp.uncertainty, i),
                             mask=this_row(inp.mask, i),
                             meta=inp.meta)
                    for i in range(inp.flux.shape[0])]

        elif isinstance(inp, (SpectrumList, SpectrumCollection)):
            return itertools.chain(*[self.input_to_list_of_spec(spec) for spec in inp])

        else:
            raise NotImplementedError(f"{inp} is not supported")

    @property
    def output(self):
        if not self.is_valid:  # pragma: nocover
            return None
        if self.input_in_sb and self.convert_to_flux_density:
            return [spectrum_ensure_flux_density_unit(sp) for sp in self.sources.selected_obj]
        else:
            return self.sources.selected_obj

    @staticmethod
    def is_wfssmulti(spec):
        return 'WFSSMulti' in spec.meta.get('header', {}).get('DATAMODL', '')

    @staticmethod
    def _extract_exposure_sourceid(spec):
        """
        Generate a label for WFSSMulti sources based on the header information.
        """
        header = spec.meta.get('header', {})
        exp_num = header.get('EXPGRPID', '0_0_0').split('_')[-2]
        source_id = str(spec.meta.get('source_id', ''))

        return exp_num, source_id

    @staticmethod
    def _has_mask(spec):
        if hasattr(spec, 'mask'):
            if spec.mask is not None and len(spec.mask):
                return True
        return False

    def _is_fully_masked(self, spec):
        if self._has_mask(spec):
            # all == True implies the entire array is masked and unusable
            if all(spec.mask):
                return True
        return False

    @property
    def _is_2d_spectrum(self):
        return isinstance(self.input, Spectrum) and self.input.flux.ndim == 2

    def _apply_spectral_mask(self, spec):
        # The masks (spec.spectral_axis.mask and spec.mask) for WFSS L3 spectra
        # may not be equivalent, so we only apply the spectral_axis mask to avoid
        # a Specutils error. Specutils expects the spectral axis to be strictly
        # increasing/decreasing so applying the 'full' mask may throw that error.
        if self._has_mask(spec.spectral_axis) and not self._is_fully_masked(spec):
            mask = spec.spectral_axis.mask
            # NOTE: Something breaks when the following is attempted instead
            # of the current implementation
            #
            # spec.mask = spec.mask | spec.spectral_axis.mask
            # return spec
            return Spectrum(
                spectral_axis=spec.spectral_axis[~mask],
                flux=spec.flux[~mask],
                uncertainty=spec.uncertainty[~mask],
                mask=mask[~mask],
                meta=spec.meta)

        return spec

    def assign_component_type(self, comp_id, comp, units, physical_type):
        return _spectrum_assign_component_type(comp_id, comp, units, physical_type)

    def __call__(self):
        if not self.sources.selected:
            raise ValueError("No sources selected.")

        with self.app._jdaviz_helper.batch_load():
            for spec_obj, item_dict in zip(self.output, self.sources.selected_item_list):
                data_label = f"{self.data_label_value}_{item_dict['suffix']}"
                self.add_to_data_collection(spec_obj, data_label)


def combine_lists_to_1d_spectrum(wl, fnu, dfnu, wave_units, flux_units):
    """
    Combine lists of 1D spectra into a composite `~specutils.Spectrum` object.

    Parameters
    ----------
    wl : list of `~astropy.units.Quantity`s
        Wavelength in each spectral channel
    fnu : list of `~astropy.units.Quantity`s
        Flux in each spectral channel
    dfnu : list of `~astropy.units.Quantity`s or None
        Uncertainty on each flux

    Returns
    -------
    spec : `~specutils.Spectrum`
        Composite 1D spectrum.
    """
    # COPIED FROM specviz.plugins.parsers since cannot import
    wlallarr = np.array(wl)
    fnuallarr = np.array(fnu)
    srtind = np.argsort(wlallarr)
    if dfnu is not None:
        dfnuallarr = np.array(dfnu)
        fnuallerr = dfnuallarr[srtind]
    wlall = wlallarr[srtind]
    fnuall = fnuallarr[srtind]

    # units are not being handled properly yet.
    if dfnu is not None:
        unc = StdDevUncertainty(fnuallerr * flux_units)
    else:
        unc = None

    spec = Spectrum(flux=fnuall * flux_units, spectral_axis=wlall * wave_units,
                    uncertainty=unc)
    return spec


@loader_importer_registry('1D Spectrum Concatenated')
class SpectrumListConcatenatedImporter(SpectrumListImporter):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.select_all_for_concatenation = False

        if self._is_2d_spectrum:
            self.disable_dropdown = True
            # If we select_all() here, then if the user switches back to Spectrum List
            # all items will be selected, which is not the desired behavior.
            self.select_all_for_concatenation = True
            # Enable the import button upon initialization because otherwise having
            # no sources selected will disable it for other valid importers (e.g. Image).
            self.import_disabled = False

    @property
    def output(self):
        if self.select_all_for_concatenation:
            self.sources.select_all()

        spectrum_list = self.sources.selected
        if len(spectrum_list) == 0:
            return []

        # Vectorized collection of all wavelengths, fluxes, and uncertainties
        wl_list = []
        fnu_list = []
        dfnu_list = []
        for spec in self.sources.selected_obj:
            wl = spec.spectral_axis.value
            fnu = spec.flux.value

            if spec.uncertainty is not None:
                dfnu = spec.uncertainty.array
            else:
                dfnu = np.full_like(fnu, np.nan)

            wl_list.append(wl)
            fnu_list.append(fnu)
            dfnu_list.append(dfnu)

        wlallorig = np.concatenate(wl_list)
        fnuallorig = np.concatenate(fnu_list)
        dfnuallorig = np.concatenate(dfnu_list)

        wave_units = spec.spectral_axis.unit
        pixar_sr = getattr(spec, 'meta', {}).get('PIXAR_SR', 1.0)
        flux_units = to_flux_density_unit(spec.flux.unit, pixar_sr)

        return combine_lists_to_1d_spectrum(wlallorig,
                                            fnuallorig,
                                            dfnuallorig,
                                            wave_units,
                                            flux_units)

    def __call__(self):
        data_label = self.data_label_value
        self.add_to_data_collection(self.output, f"{data_label}")

        # Do we need to reset in case user switches back to Spectrum List?
        # self.sources.selected = []



import os
import warnings

import pytest

from jdaviz.utils import (alpha_index, download_uri_to_path,
                          get_cloud_fits, cached_uri, escape_brackets,
                          has_wildcard, wildcard_match)
from astropy.io import fits

from jdaviz.conftest import FakeSpectrumListImporter


@pytest.mark.parametrize("test_input,expected", [(0, 'a'), (1, 'b'), (25, 'z'), (26, 'aa'),
                                                 (701, 'zz'), (702, '{a')])
def test_alpha_index(test_input, expected):
    assert alpha_index(test_input) == expected


def test_alpha_index_exceptions():
    with pytest.raises(TypeError, match="index must be an integer"):
        alpha_index(4.2)
    with pytest.raises(ValueError, match="index must be positive"):
        alpha_index(-1)


def test_uri_to_download_bad_scheme(imviz_helper):
    uri = "file://path/to/file.fits"
    with pytest.raises(ValueError, match="no valid loaders found for input"):
        imviz_helper.load_data(uri)


@pytest.mark.remote_data
def test_uri_to_download_nonexistent_mast_file(imviz_helper):
    # this validates as a mast uri but doesn't actually exist on mast:
    uri = "mast:JWST/product/jw00000-no-file-here.fits"
    with pytest.raises(ValueError, match='Failed query for URI'):
        # NOTE: this test will attempt to reach out to MAST via astroquery
        # even if cache is available.
        imviz_helper.load_data(uri, cache=False)


@pytest.mark.remote_data
def test_url_to_download_imviz_local_path_warning(imviz_helper):
    url = "https://www.astropy.org/astropy-data/tutorials/FITS-images/HorseHead.fits"

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        imviz_helper.load_data(url, cache=True, local_path='horsehead.fits')


def test_uri_to_download_specviz_local_path_check():
    # NOTE: do not use cached_uri here since no download should occur
    uri = "mast:JWST/product/jw02732-c1001_t004_miri_ch1-short_x1d.fits"
    local_path = download_uri_to_path(uri, cache=False, dryrun=True)  # No download

    # Wrong: '.\\JWST/product/jw02732-c1001_t004_miri_ch1-short_x1d.fits'
    # Correct:  '.\\jw02732-c1001_t004_miri_ch1-short_x1d.fits'
    assert local_path == os.path.join(os.curdir, "jw02732-c1001_t004_miri_ch1-short_x1d.fits")  # noqa: E501


@pytest.mark.remote_data
def test_uri_to_download_specviz(specviz_helper):
    uri = cached_uri("mast:JWST/product/jw02732-c1001_t004_miri_ch1-short_x1d.fits")
    specviz_helper.load_data(uri, cache=True)


@pytest.mark.remote_data
def test_uri_to_download_specviz2d(specviz2d_helper):
    uri = cached_uri("mast:jwst/product/jw01538-o161_t002-s000000001_nirspec_f290lp-g395h-s1600a1_s2d.fits")  # noqa: E501
    specviz2d_helper.load_data(uri, cache=True)


@pytest.mark.remote_data
def test_load_s3_fits(imviz_helper):
    """Test loading a JWST FITS file from an S3 URI into Imviz."""
    s3_uri = "s3://stpubdata/jwst/public/jw02727/L3/t/o002/jw02727-o002_t062_nircam_clear-f277w_i2d.fits"  # noqa: E501
    imviz_helper.load_data(s3_uri)
    assert len(imviz_helper.app.data_collection) > 0


@pytest.mark.remote_data
def test_get_cloud_fits_ext():
    s3_uri = "s3://stpubdata/jwst/public/jw02727/L3/t/o002/jw02727-o002_t062_nircam_clear-f277w_i2d.fits"  # noqa: E501
    hdul = get_cloud_fits(s3_uri)
    assert isinstance(hdul, fits.HDUList)

    hdul = get_cloud_fits(s3_uri, ext="SCI")
    assert isinstance(hdul, fits.HDUList)

    hdul = get_cloud_fits(s3_uri, ext=["SCI"])
    assert isinstance(hdul, fits.HDUList)


class FakeObject:
    def __init__(self):
        pass


def test_escape_brackets():
    # Test that wildcard_match escapes brackets properly
    assert escape_brackets("") == ""
    assert escape_brackets("no brackets") == "no brackets"
    assert escape_brackets("some [brackets]") == "some [[]brackets[]]"
    assert escape_brackets("s[o]me [br]a[]ckets]") == "s[[]o[]]me [[]br[]]a[[][]]ckets[]]"


def test_has_wildcard():
    assert has_wildcard("") is False
    assert has_wildcard("no wildcards") is False
    assert has_wildcard("some * wildcard") is True
    assert has_wildcard("some ? wildcard") is True
    # Don't check for brackets anymore, just * and ?
    assert has_wildcard("some [brackets] wildcard") is False
    assert has_wildcard("some [brackets] and * wildcard") is True
    assert has_wildcard("some [brackets] and ? wildcard") is True
    assert has_wildcard("*") is True
    assert has_wildcard("?") is True


def test_wildcard_match_basic(deconfigged_helper, premade_spectrum_list):
    default_choices = ['some choice', 'some good choice', 'good choice', 'maybe a bad choice']
    test_obj = FakeObject()

    # No choices in obj or provided
    match_result = wildcard_match(test_obj, '*')
    assert match_result == '*'

    # Multiselect is not an attribute yet so we retain the value of the input string
    match_result = wildcard_match(test_obj, '*', choices=default_choices[0])
    assert match_result == '*'

    # Turn on allow_multiselect to check the actual matching functionality
    test_obj.allow_multiselect = True

    # Add choices attribute, no matches
    test_obj.choices = default_choices
    match_result = wildcard_match(test_obj, 'not good*')
    assert match_result == ['not good*']

    # Test that the function kwarg `choices` overrides the object's choices attribute
    match_result = wildcard_match(test_obj, 'bad*', choices=default_choices[:-1])
    assert match_result == ['bad*']

    test_selections = {
        # Test all
        '*': default_choices,
        # Test repeats
        ('*', '* good *'): default_choices,
        # Test single selection
        'some*': default_choices[:2],
        # Test multi-wildcard
        ('*', 'good*'): default_choices,
        # Test multi-selection
        ('some*', 'good*'): default_choices[:-1]}

    for selection, expected in test_selections.items():
        # Set and reset
        test_obj.multiselect = False
        match_result = wildcard_match(test_obj, selection)
        assert test_obj.multiselect is True
        assert match_result == expected

    # Making sure a stand-in for a SelectPluginComponent object with an attribute
    # that has `choices` works as expected
    fake_importer = FakeSpectrumListImporter(app=deconfigged_helper.app,
                                             resolver=deconfigged_helper.loaders['object']._obj,
                                             input=premade_spectrum_list)
    test_obj = fake_importer.sources

    """
    Left here for reference, premade_spectrum_list has 5 spectra:
    default_choices = ['1D Spectrum at index: 0',
                       '1D Spectrum at index: 1',
                       'Exposure 0, Source ID: 0000',
                       'Exposure 0, Source ID: 1111',
                       'Exposure 1, Source ID: 1111']
    """

    test_selections = {
        # Test all
        '*': test_obj.choices,
        # Test repeats
        ('*', '*:*'): test_obj.choices,
        # Test single selection
        '1D Spectrum at index: ?': test_obj.choices[:2],
        # Test multi-wildcard
        '*Exposure*': test_obj.choices[2:],
        # Test multi-selection
        ('*at index: 1', 'Exposure 0*'): test_obj.choices[1:-1]}

    for selection, expected in test_selections.items():
        # Reset
        test_obj.multiselect = False
        match_result = wildcard_match(test_obj, selection)
        assert test_obj.multiselect is True
        assert match_result == expected
